#include <stdio.h>
#include <string.h>
#include <string>

#include "linkedlist.h"

using namespace std;

#define BLOCKSIZE 10
#define MAX_CHANGEFACTOR 15
#define RANDOM_SKIP 1 // the higher, the more are skipped.  1=don't skip any

enum colors{none,r,g,b,c,y,m,k,w};
char colorchar[9]={'\0','r','g','b','c','y','m','k','w'};

class state
{
public:
	bool EM;
	int B,U,I,PL,S,TT;
	int size;
	colors color;

	void SetAttr(char att,int val);
	int GetAttr(char att);
	bool operator==(state &other);
	void Reset(); // set everything to root values
};

// this structure is pushed into the plstack stack
typedef struct
{
	bool EM;
	int B,U,I,S,TT;
} PLattribs;

FILE *InFile(char *s)
{
	FILE *f;
	f = fopen(s,"rt");
	if (!f) {
		printf("Could not open file \"%s\"\n",s);
		return 0;
	} else
		return f;
}

typedef struct
{
	string text;
	state st;
} block;

class Path {
public:
	state current_state;

	LinkedList<colors> colorstack;
	LinkedList<int> sizestack;
	LinkedList<PLattribs> plstack;
	LinkedList<bool> emstack;

	LinkedList<char> tagstack;

	string text; // the output

	Node<block> *node;

	Path();
	void CopyFrom(Path *p);
	void OpenTag(char tag);
	char CloseTag();
	int Recurse(int level);
	void DoDefault(char att);
};

state current_state,prev_state,last_whitespace_state,prev_state_here,stored_state,empty_state;

LinkedList<colors> colorstack;
LinkedList<int> sizestack;
LinkedList<PLattribs> plstack;
LinkedList<bool> emstack;

LinkedList<block> blocks;
Node<block> *endingnode=NULL;

// for output
LinkedList<char> tagstack; // contains the first character of the tag

PLattribs temp_attribs;
bool prev_whitespace=false; // was the last character whitespace?

static string collecting_text; // used in input

int best_len;
string best_text;

string original_text;
int changefactor=0; // how much change has gone on, in the current snippet

Path mainpath;

FILE *fileo=NULL;

bool the_end=false;

void OutputSML2()
{
	mainpath.node = blocks.root;

	block b;
//	printf("trying: %s\n",original_text.c_str());

	best_len=-1;
	mainpath.Recurse(0);
	printf("best  : %s\n",best_text.c_str());
//	if (best_len==-1 || original_text.length() < best_text.length()) {
//		fprintf(fileo,"%s",original_text.c_str());
//	} else {
		fprintf(fileo,"%s",best_text.c_str());
//	}
	original_text="";
}

void WriteBlock(state *st,string *text)
{
	block bl;
	if (text) bl.text = *text;
	memcpy(&bl.st,st,sizeof(state));
	blocks.AddNode(bl);

	changefactor++;
	if (prev_state.EM != current_state.EM) changefactor++;
	if (prev_state.B != current_state.B) changefactor++;
	if (prev_state.U != current_state.U) changefactor+=abs(prev_state.U-current_state.U);
	if (prev_state.I != current_state.I) changefactor++;
	if (prev_state.S != current_state.S) changefactor++;
	if (prev_state.TT != current_state.TT) changefactor++;
	if (prev_state.size != current_state.size) changefactor++;
	if (prev_state.color != current_state.color) changefactor++;
	memcpy(&prev_state,&current_state,sizeof(state));

//	if (blocks.count >= BLOCKSIZE) {
	if (changefactor >= MAX_CHANGEFACTOR) {
		changefactor=0;
		OutputSML2();
		blocks.DeleteAll();
	}
}

void addc(char chr)
{
	bool state_changed = prev_state_here==current_state?false:true;

	// make sure to write a block, if it's the last message.
	if (chr==0) state_changed=true;

	if (state_changed) {
		if (collecting_text!="" || chr==0) {
			WriteBlock(&stored_state,&collecting_text);
		}
		collecting_text="";
		memcpy(&stored_state,&current_state,sizeof(state));
	}
	collecting_text += chr;
	if (state_changed) memcpy(&prev_state_here,&current_state,sizeof(state));
}

void state::SetAttr(char att,int val)
{
	switch (att)
	{
	case 'B': B = val; break;
	case 'E': EM = val?true:false; break;
	case 'U': U = val; break;
	case 'I': I = val; break;
	case 'P': PL = val; break;
	case 'S': S = val; break;
	case 'T': TT = val; break;
	case 'c': color = (colors)val; break;
	case 's': size = val; break;
	}
}

int state::GetAttr(char att)
{
	switch (att)
	{
	case 'B': return B;
	case 'E': return EM;
	case 'U': return U;
	case 'I': return I;
	case 'P': return PL;
	case 'S': return S;
	case 'T': return TT;
	case 'c': return color;
	case 's': return size;
	}
	return 0;
}

bool state::operator==(state &other)
{
//	return !memcmp(this,&other,sizeof(state));
	return (((EM && other.EM) || (!EM && !other.EM)) &&
			((I && other.I) || (!I && !other.I)) &&
			((B && other.B) || (!B && !other.B)) &&
			((S && other.S) || (!S && !other.S)) &&
			((TT && other.TT) || (!TT && !other.TT)) &&
			((U>3?3:U)==(other.U>3?3:other.U)) &&
			(size==other.size) &&
			(color==other.color));

}

void state::Reset()
{
	EM=false;
	B=0;
	U=0;
	I=0;
	PL=0;
	S=0;
	TT=0;
	size=-1;
	color=none;
}

Path::Path()
{
}

void Path::CopyFrom(Path *p)
{
	node = p->node;
	memcpy(&current_state,&p->current_state,sizeof(state));
	text = p->text;

	tagstack = p->tagstack;
	colorstack = p->colorstack;
	sizestack = p->sizestack;
	plstack = p->plstack;
	emstack = p->emstack;
}

void Path::OpenTag(char tag)
{
 	tagstack.Push(tag);
	text+='<';
	text+=tag;
	if (tag=='E') text+='M';
	else if (tag=='P') text+='L';
	else if (tag=='T') text+='T';
	text+='>';

	// take care of states
	switch (tag)
	{
		case 'B': current_state.B++; break;
		case 'E': if (!current_state.S) { current_state.EM = !current_state.EM; } break;
		case 'I': current_state.I++; break;
		case 'P':
			temp_attribs.EM = current_state.EM;
			temp_attribs.B = current_state.B;
			temp_attribs.U = current_state.U;
			temp_attribs.I = current_state.I;
			temp_attribs.S = current_state.S;
			temp_attribs.TT = current_state.TT;
			plstack.Push(temp_attribs);
			current_state.EM=false;
			current_state.B=0;
			current_state.U=0;
			current_state.I=0;
			current_state.S=0;
			current_state.TT=0;
			current_state.PL++;
			break;
		case 'S': emstack.Push(current_state.EM); current_state.EM=false; current_state.S++; break;
		case 'T': current_state.TT++; break;
		case 'U': current_state.U++; break;
		case '0': sizestack.Push(current_state.size); current_state.size=0; break;
		case '1': sizestack.Push(current_state.size); current_state.size=1; break;
		case '2': sizestack.Push(current_state.size); current_state.size=2; break;
		case '3': sizestack.Push(current_state.size); current_state.size=3; break;
		case '4': sizestack.Push(current_state.size); current_state.size=4; break;
		case '5': sizestack.Push(current_state.size); current_state.size=5; break;
		case '6': sizestack.Push(current_state.size); current_state.size=6; break;
		case '7': sizestack.Push(current_state.size); current_state.size=7; break;
		case '8': sizestack.Push(current_state.size); current_state.size=8; break;
		case '9': sizestack.Push(current_state.size); current_state.size=9; break;
		case 'r': colorstack.Push(current_state.color); current_state.color=r; break;
		case 'g': colorstack.Push(current_state.color); current_state.color=g; break;
		case 'b': colorstack.Push(current_state.color); current_state.color=b; break;
		case 'c': colorstack.Push(current_state.color); current_state.color=c; break;
		case 'y': colorstack.Push(current_state.color); current_state.color=y; break;
		case 'm': colorstack.Push(current_state.color); current_state.color=m; break;
		case 'k': colorstack.Push(current_state.color); current_state.color=k; break;
		case 'w': colorstack.Push(current_state.color); current_state.color=w; break;
	}
}

char Path::CloseTag()
{
	char tag;
	tag = tagstack.Pop();
	text+="</";
	text+=tag;
	if (tag=='E') text+='M';
	else if (tag=='P') text+='L';
	else if (tag=='T') text+='T';
	text+='>';

	// take care of states
	switch (tag)
	{
		case 'B': current_state.B--; break;
		case 'E': current_state.EM = !current_state.EM; break;
		case 'I': current_state.I--; break;
		case 'P':
			temp_attribs = plstack.Pop();
			current_state.EM = temp_attribs.EM;
			current_state.B = temp_attribs.B;
			current_state.U = temp_attribs.U;
			current_state.I = temp_attribs.I;
			current_state.S = temp_attribs.S;
			current_state.TT = temp_attribs.TT;
			current_state.PL--;
			break;
		case 'S': current_state.EM = emstack.Pop(); current_state.S--; break;
		case 'T': current_state.TT--; break;
		case 'U': current_state.U--; break;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': current_state.size = sizestack.Pop(); break;
		case 'r':
		case 'g':
		case 'b':
		case 'c':
		case 'y':
		case 'm':
		case 'k':
		case 'w': current_state.color = colorstack.Pop(); break;
	}

	return tag;
}

#define NEW_PATH(from)	Path newpath; \
						newpath.CopyFrom(from);

#define RUN_PATH		{newpath.text += bl.text; \
						newpath.node = newpath.node->next; \
						if (bl.st==newpath.current_state) newpath.Recurse(0);}

#define RUN_ALL_ATTRIBUTES(path_) \
	if (att!='S' && !(rand()%RANDOM_SKIP)) path_.DoDefault('S'); \
	if (att!='E' && !(rand()%RANDOM_SKIP)) path_.DoDefault('E'); \
	if (att!='I' && !(rand()%RANDOM_SKIP)) path_.DoDefault('I'); \
	if (att!='B' && !(rand()%RANDOM_SKIP)) path_.DoDefault('B'); \
	if (            !(rand()%RANDOM_SKIP)) path_.DoDefault('U'); \
	if (att!='T' && !(rand()%RANDOM_SKIP)) path_.DoDefault('T'); \
	if (att!='c' && !(rand()%RANDOM_SKIP)) path_.DoDefault('c'); \
	if (att!='s' && !(rand()%RANDOM_SKIP)) path_.DoDefault('s'); \
	if (att!='P' && !(rand()%RANDOM_SKIP)) path_.DoDefault('P');

#define DO_SOME_PL_STUFF(from) \
	{ NEW_PATH(from); \
	newpath.OpenTag('P'); \
	if (newpath.current_state==bl.st) { \
		RUN_PATH; \
	} else { \
		/*RUN_ALL_ATTRIBUTES(newpath);*/ \
	}}

void Path::DoDefault(char att)
{
//	if (att=='P') return;

	block bl = node->data;

	if (att=='P') {
		if (current_state.PL) {
			NEW_PATH(this);
			char a;
			do {
				a = newpath.CloseTag();
			} while (a!='P' && !newpath.tagstack.Empty());
			if (a=='P') RUN_PATH;
//			RUN_ALL_ATTRIBUTES(newpath);
		}
	
		if (bl.st==empty_state) {
			NEW_PATH(this);
			newpath.OpenTag('P');
			RUN_PATH;
		}
		return;
	}

	// U
	if (att == 'U' && bl.st.U != current_state.U) {
		// if i need to set it
		if (current_state.U < bl.st.U) {
			NEW_PATH(this);
			newpath.OpenTag('U');
			if (newpath.current_state==bl.st) {
				RUN_PATH;
			} else {
				RUN_ALL_ATTRIBUTES(newpath);
			}
		}

		// if i need to un-set it
		if (current_state.U > bl.st.U) {
			Path testpath;
			testpath.CopyFrom(this);
			NEW_PATH(this);
			while (!testpath.tagstack.Empty()) {
				DO_SOME_PL_STUFF(&testpath);
				testpath.CloseTag();
				if (testpath.current_state.U == bl.st.U) {
					if (testpath.current_state==bl.st) {
						NEW_PATH(&testpath);
						RUN_PATH;
					} else {
						RUN_ALL_ATTRIBUTES(testpath);
					}
				}
			}
			DO_SOME_PL_STUFF(&testpath);
		}
		return;
	}

	// color
	if (att=='c' && bl.st.color != current_state.color) {
		// try opening the tag
		if (bl.st.color != none) {
			NEW_PATH(this);
			newpath.OpenTag(colorchar[bl.st.color]);
			if (newpath.current_state==bl.st) {
				RUN_PATH;
			} else {
				RUN_ALL_ATTRIBUTES(newpath);
			}
		}
		// try closing tags
		Path testpath;
		testpath.CopyFrom(this);
		while (!testpath.tagstack.Empty()) {
			DO_SOME_PL_STUFF(&testpath);
			testpath.CloseTag();
			if (testpath.current_state.color == bl.st.color) {
				if (testpath.current_state==bl.st) {
					NEW_PATH(&testpath);
					RUN_PATH;
				} else {
					RUN_ALL_ATTRIBUTES(testpath);
				}
			}
		}
		DO_SOME_PL_STUFF(&testpath);
		return;
	}

	// size
	if (att=='s' && bl.st.size != current_state.size) {
		{
			NEW_PATH(this);
			newpath.OpenTag('0'+bl.st.size);
			if (newpath.current_state==bl.st) {
				RUN_PATH;
			} else {
				RUN_ALL_ATTRIBUTES(newpath);
			}
		}
		Path testpath;
		testpath.CopyFrom(this);
		while (!testpath.tagstack.Empty()) {
			DO_SOME_PL_STUFF(&testpath);
			testpath.CloseTag();
			if (testpath.current_state.size == bl.st.size) {
				if (testpath.current_state==bl.st) {
					NEW_PATH(&testpath);
					RUN_PATH;
				} else {
					RUN_ALL_ATTRIBUTES(testpath);
				}
			}
		}
		DO_SOME_PL_STUFF(&testpath);
		return;
	}

	// if i need to set it
	if (bl.st.GetAttr(att) && !current_state.GetAttr(att)) {
		NEW_PATH(this);
		newpath.OpenTag(att);
		if (newpath.current_state==bl.st) {
			RUN_PATH;
		} else {
			RUN_ALL_ATTRIBUTES(newpath);
		}
	}

	// if i need to un-set it
	if (!bl.st.GetAttr(att) && current_state.GetAttr(att)) {
		Path testpath;
		testpath.CopyFrom(this);
		NEW_PATH(this);
		while (!testpath.tagstack.Empty()) {
			DO_SOME_PL_STUFF(&testpath);
			testpath.CloseTag();
			if (!testpath.current_state.GetAttr(att)) {
				if (testpath.current_state==bl.st) {
					NEW_PATH(&testpath);
					RUN_PATH;
				} else {
					RUN_ALL_ATTRIBUTES(testpath);
				}
			}
		}
		DO_SOME_PL_STUFF(&testpath);
	}
}

int Path::Recurse(int level)
{
//	printf("hey %s\n",text.c_str());
	if (!node || node==endingnode) {
//		if (!node) while (!tagstack.Empty()) CloseTag();
		if (the_end) while (!tagstack.Empty()) CloseTag();
		int len=text.length();
//		if (text.find("<PL> bar ")!=-1) printf("found: %s %d\n",text.c_str(),text.find("<PL> bar </PL>"));
		if (original_text==" </4><U> a</U><S> </S>  </EM> a") {
			static int a=0;
			a++;
//			if (!(a%1000)) printf("found: %s\n",text.c_str());
		}
		if (len<best_len || best_len==-1) {
			best_len = len;
			best_text = text;
		}
		return 0;
	}

	block bl = node->data;

	if (bl.st==current_state) {
		NEW_PATH(this);
		RUN_PATH;
		return 0;
	}

	DoDefault('B');
	DoDefault('I');
	DoDefault('T');
	DoDefault('S');
	DoDefault('U');
	DoDefault('E');
	DoDefault('c');
	DoDefault('s');
	DoDefault('P');

	return 0;
}

void OutputSML()
{
	int i;

	endingnode = blocks.root;

	Path mainpath;
	mainpath.node = blocks.root;
	mainpath.current_state.Reset();

	do {

		for (i=0; i<BLOCKSIZE && endingnode; i++)
			endingnode = endingnode->next;

		block b;
		printf("trying: ");
		Node<block> *n = mainpath.node;
		while (n!=endingnode) {
			b = n->data;
			if (b.st.B) printf("<B>");
			if (b.st.color) printf("<%c>",colorchar[b.st.color]);
			if (b.st.EM) printf("<EM>");
			if (b.st.I) printf("<I>");
			if (b.st.S) printf("<S>");
			if (b.st.size!=-1) printf("<%c>",'0'+b.st.size);
			if (b.st.TT) printf("<TT>");
			if (b.st.U) printf("<U>");
			printf("%s",b.text.c_str(),b.st.U);
			if (b.st.B) printf("</B>");
			if (b.st.color) printf("</%c>",colorchar[b.st.color]);
			if (b.st.EM) printf("</EM>");
			if (b.st.I) printf("</I>");
			if (b.st.S) printf("</S>");
			if (b.st.size!=-1) printf("</%c>",'0'+b.st.size);
			if (b.st.TT) printf("</TT>");
			if (b.st.U) printf("</U>");
			n=n->next;
		}
		printf("\n");

		best_len=-1;
		mainpath.Recurse(0);
		printf("best = %s\n",best_text.c_str());
		mainpath.node = endingnode;

	} while (endingnode);
}

void main()
{
	srand(35656);
//#define CR
#ifdef CR
	FILE *f = fopen("test.txt","wt");
	string ts;
	for (long i=0; i<100; i++) {
//		printf("%d\n",tagstack.count);
	if (tagstack.Empty() || rand()%2 && tagstack.count<5) {
		int t=rand()%25;
		tagstack.Push(t);
		switch (t)
		{
		case 0: ts="B"; break;
		case 1: ts="EM"; break;
		case 2: ts="I"; break;
		case 3: ts="PL"; break;
		case 4: ts="S"; break;
		case 5: ts="TT"; break;
		case 6: ts="U"; break;
		case 7: ts="0"; break;
		case 8: ts="1"; break;
		case 9: ts="2"; break;
		case 10: ts="3"; break;
		case 11: ts="4"; break;
		case 12: ts="5"; break;
		case 13: ts="6"; break;
		case 14: ts="7"; break;
		case 15: ts="8"; break;
		case 16: ts="9"; break;
		case 17: ts="r"; break;
		case 18: ts="g"; break;
		case 19: ts="b"; break;
		case 20: ts="c"; break;
		case 21: ts="y"; break;
		case 22: ts="m"; break;
		case 23: ts="k"; break;
		case 24: ts="w"; break;
		}
		fprintf(f,"<%s>",ts.c_str());
		if (rand()%2) fprintf(f," ");
		if (rand()%2) fprintf(f,"a");
		if (rand()%2) fprintf(f," ");
	} else {
		int t=tagstack.Pop();
		switch (t)
		{
		case 0: ts="B"; break;
		case 1: ts="EM"; break;
		case 2: ts="I"; break;
		case 3: ts="PL"; break;
		case 4: ts="S"; break;
		case 5: ts="TT"; break;
		case 6: ts="U"; break;
		case 7: ts="0"; break;
		case 8: ts="1"; break;
		case 9: ts="2"; break;
		case 10: ts="3"; break;
		case 11: ts="4"; break;
		case 12: ts="5"; break;
		case 13: ts="6"; break;
		case 14: ts="7"; break;
		case 15: ts="8"; break;
		case 16: ts="9"; break;
		case 17: ts="r"; break;
		case 18: ts="g"; break;
		case 19: ts="b"; break;
		case 20: ts="c"; break;
		case 21: ts="y"; break;
		case 22: ts="m"; break;
		case 23: ts="k"; break;
		case 24: ts="w"; break;
		}
		fprintf(f,"</%s>",ts.c_str());
		if (rand()%2) fprintf(f," ");
		if (rand()%2) fprintf(f,"a");
		if (rand()%2) fprintf(f," ");
	}
	}
	while (!tagstack.Empty()) {
		int t=tagstack.Pop();
		switch (t)
		{
		case 0: ts="B"; break;
		case 1: ts="EM"; break;
		case 2: ts="I"; break;
		case 3: ts="PL"; break;
		case 4: ts="S"; break;
		case 5: ts="TT"; break;
		case 6: ts="U"; break;
		case 7: ts="0"; break;
		case 8: ts="1"; break;
		case 9: ts="2"; break;
		case 10: ts="3"; break;
		case 11: ts="4"; break;
		case 12: ts="5"; break;
		case 13: ts="6"; break;
		case 14: ts="7"; break;
		case 15: ts="8"; break;
		case 16: ts="9"; break;
		case 17: ts="r"; break;
		case 18: ts="g"; break;
		case 19: ts="b"; break;
		case 20: ts="c"; break;
		case 21: ts="y"; break;
		case 22: ts="m"; break;
		case 23: ts="k"; break;
		case 24: ts="w"; break;
		}
		fprintf(f,"</%s>",ts.c_str());
		if (rand()%2) fprintf(f,"a");
	}
	fclose(f);
	return;
#endif

	FILE *file;
	char chr;
	char tag[4]="\0\0\0";
	int tag_size=0;

	fileo = fopen("output.txt","wt");
//	fileo = stdout;

	current_state.Reset();
	stored_state.Reset();
	empty_state.Reset();
	prev_state.Reset();
	prev_state_here.Reset();

	mainpath.current_state.Reset();

	file = InFile("infile.txt");
//	file = stdin;
	while (file && !feof(file))
	{
		chr=fgetc(file);
		original_text += chr;
		if (chr=='<')
		{
			tag_size=0;
			do {
				chr=fgetc(file);
				original_text += chr;
				if (tag_size<3) tag[tag_size++]=chr;
			} while (chr!='>' && !feof(file));
			switch(tag[0])
			{
				case 'B': current_state.B++; break;
				case 'E': if (!current_state.S) { current_state.EM = !current_state.EM; } break;
				case 'I': current_state.I++; break;
				case 'P':
					temp_attribs.EM = current_state.EM;
					temp_attribs.B = current_state.B;
					temp_attribs.U = current_state.U;
					temp_attribs.I = current_state.I;
					temp_attribs.S = current_state.S;
					temp_attribs.TT = current_state.TT;
					plstack.Push(temp_attribs);
					current_state.EM=false;
					current_state.B=0;
					current_state.U=0;
					current_state.I=0;
					current_state.S=0;
					current_state.TT=0;
					current_state.PL++;
					break;
				case 'S': emstack.Push(current_state.EM); current_state.EM=false; current_state.S++; break;
				case 'T': current_state.TT++; break;
				case 'U': current_state.U++; break;
				case '0': sizestack.Push(current_state.size); current_state.size=0; break;
				case '1': sizestack.Push(current_state.size); current_state.size=1; break;
				case '2': sizestack.Push(current_state.size); current_state.size=2; break;
				case '3': sizestack.Push(current_state.size); current_state.size=3; break;
				case '4': sizestack.Push(current_state.size); current_state.size=4; break;
				case '5': sizestack.Push(current_state.size); current_state.size=5; break;
				case '6': sizestack.Push(current_state.size); current_state.size=6; break;
				case '7': sizestack.Push(current_state.size); current_state.size=7; break;
				case '8': sizestack.Push(current_state.size); current_state.size=8; break;
				case '9': sizestack.Push(current_state.size); current_state.size=9; break;
				case 'r': colorstack.Push(current_state.color); current_state.color=r; break;
				case 'g': colorstack.Push(current_state.color); current_state.color=g; break;
				case 'b': colorstack.Push(current_state.color); current_state.color=b; break;
				case 'c': colorstack.Push(current_state.color); current_state.color=c; break;
				case 'y': colorstack.Push(current_state.color); current_state.color=y; break;
				case 'm': colorstack.Push(current_state.color); current_state.color=m; break;
				case 'k': colorstack.Push(current_state.color); current_state.color=k; break;
				case 'w': colorstack.Push(current_state.color); current_state.color=w; break;
				case '/':
				switch(tag[1])
				{
					case 'B': current_state.B--; break;
					case 'E': current_state.EM = !current_state.EM; break;
					case 'I': current_state.I--; break;
					case 'P':
						temp_attribs = plstack.Pop();
						current_state.EM = temp_attribs.EM;
						current_state.B = temp_attribs.B;
						current_state.U = temp_attribs.U;
						current_state.I = temp_attribs.I;
						current_state.S = temp_attribs.S;
						current_state.TT = temp_attribs.TT;
						current_state.PL--;
						break;
					case 'S': current_state.EM = emstack.Pop(); current_state.S--; break;
					case 'T': current_state.TT--; break;
					case 'U': current_state.U--; break;
					case '0':
					case '1':
					case '2':
					case '3':
					case '4':
					case '5':
					case '6':
					case '7':
					case '8':
					case '9': current_state.size = sizestack.Pop(); break;
					case 'r':
					case 'g':
					case 'b':
					case 'c':
					case 'y':
					case 'm':
					case 'k':
					case 'w': current_state.color = colorstack.Pop(); break;
				}
			}
		} else if (chr>0) {
			bool whitespace = (chr=='\n' || chr=='\r' || chr=='\t' || chr==' ');
			if (whitespace && !current_state.TT) {
				if (!prev_whitespace ||
					current_state.size!=last_whitespace_state.size ||
					(current_state.U>0 && current_state.color!=last_whitespace_state.color)) {

					addc(' ');
					memcpy(&last_whitespace_state,&current_state,sizeof(state));
				}
			} else {
				addc(chr);
				memcpy(&last_whitespace_state,&current_state,sizeof(state));
			}
			prev_whitespace = whitespace;
		}
	}
	WriteBlock(&stored_state,&collecting_text);// force a block write for the last block of text.
	current_state.Reset();
//	WriteBlock(&current_state,NULL);// force a block write for the root state

//	OutputSML();

	the_end=true;
	OutputSML2();

	fclose(fileo);
}
